# tests_view_controller.rb
# RCRunner
#
# Created by Juri Pakaste on 21.11.10.
# Copyright 2010 Juri Pakaste. All rights reserved.

require 'set'

require 'cocoa_tools'
require 'tests_provider_facade'

module RCRunner
  class TestsViewController < NSViewController
    attr_writer :tests_outline, :output_view, :tests_progress, :test_suites
    attr_writer :main_split, :result_split, :error_text_view, :error_container
    attr_accessor :selection, :test_suites
    attr_accessor :run_button, :run_selected_button, :run_failed_button

    include CocoaTools

    def initWithNibName(name, bundle:bundle)
      super
      # We need to use KVO to look at selection because MacRuby doesn't allow
      # us to override the setter method.
      self.addObserver(self, forKeyPath:'selection', options:NSKeyValueObservingOptionNew, context:nil)
    end

    def loadView
      super

      priority_split_delegate = PrioritySplitViewDelegate.new
      priority_split_delegate.lengths_by_view_index = {0 => 320, 1 => 60}
      priority_split_delegate.view_indices_by_priority = {1 => 1, 2 => 0}
      @main_split.delegate = priority_split_delegate
    end
    
    def observeValueForKeyPath(keyPath, ofObject:object, change:change, context:context)
      if object == self and keyPath == 'selection'
        # Only show error if there's one selection and it's not a top
        # level node.
        self.display_result_for_selection
      else
        super
      end
    end

    def display_result_for_selection_if_node(testnode)
      if @selection.length == 1 and @selection[0].length == 2
        ip = @selection[0]
        selectednode = @test_suites[ip.indexAtPosition(0)].tests[ip.indexAtPosition(1)]
        if testnode.nil? or selectednode == testnode
          self.show_result(selectednode)
        end
      else
        self.clear_result_display
      end
    end

    def display_result_for_selection
      self.display_result_for_selection_if_node(nil)
    end

    def clear_result_display
      @output_view.textStorage.mutableString.setString("")
      @error_text_view.textStorage.mutableString.setString("")
    end
    
    def show_result(testnode)
      result = testnode.result
      if result.nil?
        self.clear_result_display
      else
        if [:error, :failure].include?(result.outcome) then
          if @error_container.frame.size.height < 5
            errframe = @error_container.frame
            errframe.size.height = 100
            @error_container.animator.frame = errframe
          end
          bt = result.backtrace.join("\n")
          fullstr = "#{result.reason}\n\n#{bt}"
          @error_text_view.textStorage.mutableString.setString(fullstr)
          @error_text_view.textStorage.removeAttribute(NSFontAttributeName, range:NSRange.new(0, fullstr.length))
          @error_text_view.textStorage.setAttributes({NSFontAttributeName => NSFont.fontWithName("Helvetica-Bold", size:14)}, range:NSRange.new(0, result.reason.length))
        else
          errframe = @error_container.frame
          errframe.size.height = 0
          @error_container.animator.frame = errframe
          @error_text_view.textStorage.mutableString.setString("")
        end
        @output_view.textStorage.mutableString.setString("#{result.log}")
      end
    end

    def run_test(testnode)
      puts "running testnode: #{testnode}"
      result = testnode.test.run
      self.willChangeValueForKey("test_suites")
      testnode.result_adapter = TestResultDisplayAdapter.new(result)
      self.didChangeValueForKey("test_suites")
      self.display_result_for_selection_if_node(testnode)
    end

    def run_test_list(tests)
      buttons = [@run_button, @run_failed_button, @run_selected_button]
      buttons.each { |button| button.enabled = false }
      @tests_progress.setDoubleValue(0)
      @tests_progress.setMaxValue(tests.length)
      self.each_tail_runloop(tests.shuffle) do |testnode, tail|
        self.run_test(testnode)
        @tests_progress.setDoubleValue(@tests_progress.maxValue - tail.length)
        if tail.length == 0
          buttons.each { |button| button.enabled = true }
          @tests_progress.setDoubleValue(0)
        end
      end
    end
    
    def run_all
      tests = @suitehash.values.flatten
      self.run_test_list(tests)
    end

    def run_clicked(sender)
      self.run_all
    end

    def run_selected_clicked(sender)
      # Array.new to force a mutable array, calling sort (or sort!,
      # obviously) causes an exception otherwise
      ips = Array.new(@selection).map do |indexPath|
        [indexPath.indexAtPosition(0),
         indexPath.length < 2 ? -1 : indexPath.indexAtPosition(1)]
      end

      ips.sort!
      tests_set = Set.new()
      tests_list = []
      ips.each do |iparr|
        suiteidx, testidx = iparr
        suite = @test_suites[suiteidx]
        if testidx == -1
          suite.tests.reject { |tn| tests_set.include? tn }.each do |testnode|
            tests_list << testnode
            tests_set << testnode
          end
        else
          testnode = suite.tests[testidx]
          if not tests_set.include? testnode
            tests_list << testnode
            tests_set << testnode
          end
        end
      end

      self.run_test_list(tests_list)
    end

    def run_failed_clicked(sender)
      failed = @suitehash.values.flatten.reject { |tn| tn.result.nil? or tn.result.outcome == :success }
      self.run_test_list(failed)
    end

    def tests_provider=(tprovider)
      @tests_provider = tprovider
      @suitehash = tprovider.tests.map {|t| TestTreeNode.new(t) }.group_by {|t| SuiteTreeNode.new(t.test.suite) }
      @suitehash.each { |suite, tests| suite.tests = tests.sort { |t1, t2| t1.name <=> t2.name } }
      self.test_suites = @suitehash.keys
      puts "test_suites: #{self.test_suites}"
      @tests_outline.expandItem(nil, expandChildren:true)
    end

  end

  
  # Priority Split View Delegate to control split view subview sizes
  # See http://cocoawithlove.com/2009/09/nssplitview-delegate-for-priority-based.html

  class PrioritySplitViewDelegate
    attr_accessor :lengths_by_view_index, :view_indices_by_priority
    
    def splitView(sender, constrainMinCoordinate:proposed_min, ofSubviewAt:offset)
      subview = sender.subviews[offset]
      subview_frame = subview.frame
      if sender.isVertical
        frame_origin = subview_frame.origin.x
      else
        frame_origin = subview_frame.origin.y
      end

      minimum_size = @lengths_by_view_index[offset]
      frame_origin + minimum_size
    end


    def splitView(sender, constrainMaxCoordinate:proposed_max, ofSubviewAt:offset)
      growing_subview = sender.subviews[offset]
      shrinking_subview = sender.subviews[offset + 1]
      growing_subview_frame = growing_subview.frame
      shrinking_subview_frame = shrinking_subview.frame

      if sender.isVertical
        current_coordinate = growing_subview_frame.origin.x + growing_subview_frame.size.width
        shrinking_size = shrinking_subview_frame.size.width
      else
        current_coordinate = growing_subview_frame.origin.y + growing_subview_frame.size.height
        shrinking_size = shrinking_subview_frame.size.height
      end

      minimum_size = @lengths_by_view_index[offset+1]
      current_coordinate + (shrinking_size - minimum_size)
    end


    def splitView(sender, resizeSubviewsWithOldSize:old_size)
      subviews = sender.subviews
      subviews_count = subviews.length
      is_vertical = sender.isVertical

      delta = (is_vertical ?
               (sender.bounds.size.width - old_size.width) :
               (sender.bounds.size.height - old_size.height))

      view_count_check = 0

      @view_indices_by_priority.keys.sort.each do |priority_index|
        view_index = @view_indices_by_priority[priority_index]
        next if view_index >= subviews_count

        view = subviews[view_index]
        frame_size = view.frame.size
        min_length = @lengths_by_view_index[view_index]

        if is_vertical
          frame_size.height = sender.bounds.size.height
          if delta > 0 or frame_size.width + delta >= min_length
            frame_size.width += delta
            delta = 0
          elsif delta < 0
            delta += frame_size.width - min_length
            frame_size.width = min_length
          end
        else
          frame_size.width = sender.bounds.size.width
          if delta > 0 or frame_size.height + delta >= min_length
            frame_size.height += delta
            delta = 0
          elsif delta < 0
            delta += frame_size.height - min_length
            frame_size.height = min_length
          end
        end

        view.setFrameSize(frame_size)
        view_count_check += 1

      end

      raise RangeError, "Number of valid views in priority list (#{view_count_check}) is less than the subview count (#{subviews_count}) of split view #{sender}" if view_count_check != subviews_count
      raise RuntimeError, "Split view #{sender} (#{sender.frame.size.width}x#{sender.frame.size.height}) resized smaller than minimum #{is_vertical ? 'width' : 'height'} of #{ sender.frame.size.width - delta }" if delta.abs > 0.5

      offset = 0
      divider_thickness = sender.dividerThickness
      subviews.each do |subview|
        view_frame = subview.frame
        view_origin = view_frame.origin
        view_origin.x = offset
        subview.setFrameOrigin(view_origin)
        offset += view_frame.size.width + divider_thickness
      end
    end
  end

  

  # Wrapper classes for displaying suites, tests and results
  
  class TreeNode
    attr_accessor :result_adapter
    
    def name() nil end
    def children() [] end
    def leaf() true end
    def result() @result_adapter.nil? ? nil : @result_adapter.result end
  end

  class TestTreeNode < TreeNode
    attr_reader :test

    def initialize(test)
      @test = test
    end

    def name() @test.name end
  end


  class SuiteTreeNode < TreeNode
    attr_reader :suite
    attr_accessor :tests

    def initialize(suite, tests = [])
      @suite = suite
      @tests = tests
    end

    def ==(ob) ob.kind_of?(SuiteTreeNode) and ob.suite == @suite end
    def eql?(ob) ob.kind_of?(SuiteTreeNode) and ob.suite == @suite end
    def name() @suite end
    def children() tests end
    def leaf() false end
    def hash() self.class.hash ^ @suite.hash end
  end

  class TestResultDisplayAdapter
    attr_reader :result
    
    def initialize(result)
      @result = result
    end

    def text() @result.outcome.to_s end

    def color
      case @result.outcome
        when :error then NSColor.colorWithDeviceRed(0.8, green: 0, blue: 0, alpha: 1)
        when :failure then NSColor.colorWithDeviceRed(1, green: 0, blue: 0, alpha: 1)
        when :success then NSColor.colorWithDeviceRed(0, green:0.8, blue: 0, alpha: 1)
        else NSColor.blackColor
      end
    end
  end
end
